
/*-------------------------------------------------*/
/* fireworks.c - graphical MIDI note visualisation */
/*           1998 by Christian Buchner            */
/*-------------------------------------------------*/

#define FPS 25

#define MAXCOORD 1024

#define CENTER_X 512		/* window width  is 0...MAXCOORD */
#define CENTER_Y 980		/* window height is 0...MAXCOORD */

#define NUMPENS 17

UBYTE PenColors[NUMPENS][3] =
{
	0x00, 0x00, 0x00,			/* Black background */
	
	255,  64,  64,			/* Chan  1 */
	255, 137,  64,			/* Chan  2 */
	255, 207,  64,			/* Chan  3 */
	230, 255,  64,			/* Chan  4 */
	159, 255,  64,			/* Chan  5 */
	 86, 255,  64,			/* Chan  6 */
	 64, 255, 112,			/* Chan  7 */
	 64, 255, 185,			/* Chan  8 */
	 64, 255, 255,			/* Chan  9 */
	 64, 185, 255,			/* Chan 10 */
	 64, 112, 255,			/* Chan 11 */
	 86,  64, 255,			/* Chan 12 */
	159,  64, 255,			/* Chan 13 */
	230,  64, 255,			/* Chan 14 */
	255,  64, 207,			/* Chan 15 */
	255,  64, 137,			/* Chan 16 */
};

enum
{
	Backgroundpen = 0,
	Channelpens = 1
};


static UWORD chip WaitPointer[] =
{
	0x0000, 0x0000,

	0x0400, 0x07C0,
	0x0000, 0x07C0,
	0x0100, 0x0380,
	0x0000, 0x07E0,
	0x07C0, 0x1FF8,
	0x1FF0, 0x3FEC,
	0x3FF8, 0x7FDE,
	0x3FF8, 0x7FBE,
	0x7FFC, 0xFF7F,
	0x7EFC, 0xFFFF,
	0x7FFC, 0xFFFF,
	0x3FF8, 0x7FFE,
	0x3FF8, 0x7FFE,
	0x1FF0, 0x3FFC,
	0x07C0, 0x1FF8,
	0x0000, 0x07E0,

	0x0000, 0x0000,		/* reserved, must be NULL */
};


/* Library bases */

struct ExecBase			*SysBase;
struct DosLibrary		*DOSBase;
struct GfxBase			*GfxBase;
struct Library			*LayersBase;
struct IntuitionBase	*IntuitionBase;
struct Library 			*GadToolsBase;
struct Library			*UtilityBase;
struct Library			*DatatypesBase;
struct Library			*AslBase;
struct Library			*CamdBase;
struct Library			*TimerBase;
BOOL					 FPBase;	/* pseudo floating point base */

/* Workbench or not? */

BOOL WBMode = FALSE;

/* this task */
struct Task *MyTask;

/* Prototypes */

LONG ShellInterface(void);
LONG WBInterface(struct Process *MyProc);

BOOL WindowLayout(struct Screen *scr, WORD *ww, WORD *wh, WORD *minw, WORD *minh, WORD *maxw, WORD *maxh);

LONG Fireworks(UBYTE *linkname, LONG WinX, LONG WinY, LONG WinW, LONG WinH, UBYTE *Image);
void MainLoop(UBYTE *linkname, struct Window *Window, WORD *ww, WORD *wh, struct MidiNode **midi, struct MidiLink **link, APTR NotePool, struct timerequest *treq, LONG *PenArray, struct BitMap **BGBitMap, struct BitMap **PaintBitMap, struct RastPort *PaintRP, struct Layer_Info **LInfo, struct Layer **PaintLayer, UBYTE *Image, Object **dto);

int __interrupt __saveds VBlankInterface( void );

BOOL InitOrUpdateGraphics(struct Window *Window, WORD ww, WORD wh, LONG *PenArray, struct BitMap **BGBitMap, struct BitMap **PaintBitMap, struct RastPort *PaintRP, struct Layer_Info **LInfo, struct Layer **PaintLayer, UBYTE *Image, Object **dto);
void FreeGraphics(struct Window *Window, UWORD ww, UWORD wh, LONG *PenArray, struct BitMap **BGBitMap, struct BitMap **PaintBitMap, struct Layer_Info **LInfo, struct Layer **PaintLayer, Object **dto);

void LoadImage(struct Window *Window, UWORD ww, UWORD wh, LONG *PenArray, UBYTE *Image, Object **dto);
BOOL SelectImage(struct Window *Window, UBYTE *filebuffer, ULONG MaxSize);

BOOL OpenLibs(void);
void CloseLibs(void);

struct timerequest *OpenTimer(void);
void CloseTimer(struct timerequest *treq);
ULONG GetTimeDelta(void);

LONG __stdargs Message(UBYTE *Msg,UBYTE *Options,...);
void __stdargs _XCEXIT(LONG lcode);

extern BOOL __stdargs __fpinit(void);
extern void __stdargs __fpterm(void);

                              /* SPrintf.a */
APTR __stdargs SPrintf(char *, const char *, ...);
APTR __stdargs VSprintf(char *, const char *, va_list);


/* CAM Library data and protos */

struct MidiNode *CreateMidi(Tag tag, ...);
BOOL SetMidiAttrs(struct MidiNode *mi, Tag tag, ...);
struct MidiLink *AddMidiLink(struct MidiNode *mi, LONG type, Tag tag, ...);
BOOL SetMidiLinkAttrs(struct MidiLink *mi, Tag tag, ...);


/* Some defines */

/* the opposite of noteon() */
#define noteoff(m) ( voicemsg(m,MS_NoteOff) || (voicemsg(m,MS_NoteOn) && (!(m)->mm_Data2)) )

#define offsetof(s, m)  (size_t)(&(((s *)0)->m))
#define MIN(a,b)    ((a) <= (b) ? (a) : (b))
#define elementsof(a)  ((sizeof(a) / sizeof(a[0])))


/* Intuition structures */

enum
{
	Menu_Ignore,
	Menu_Link,
	Menu_Save,
	Menu_About,
	Menu_Quit,
	Menu_Kill,
	Menu_Load,
	Menu_Double,
	Menu_Release,
};


struct NewMenu VUNewMenu[]=
{
	/* nm_Type	nm_Label			nm_CommKey	nm_Flags			nm_MutualExclude	nm_UserData */
	NM_TITLE,	"Project",			NULL,		0,					0,					(APTR)Menu_Ignore,
	NM_ITEM,	"MIDI Link...",		"L",		0,					0,					(APTR)Menu_Link,
	NM_ITEM,	"Save Config",		"S",		0,					0,					(APTR)Menu_Save,
	NM_ITEM,	NM_BARLABEL,		NULL,		0,					0,					(APTR)Menu_Ignore,
	NM_ITEM,	"About",			"?",		0,					0,					(APTR)Menu_About,
	NM_ITEM,	NM_BARLABEL,		NULL,		0,					0,					(APTR)Menu_Ignore,
	NM_ITEM,	"Quit",				"Q",		0,					0,					(APTR)Menu_Quit,
	NM_TITLE,	"Backdrop",			NULL,		0,					0,					(APTR)Menu_Ignore,
	NM_ITEM,	"Remove",			"R",		0,					0,					(APTR)Menu_Kill,
	NM_ITEM,	"Load Another...",	"A",		0,					0,					(APTR)Menu_Load,
	NM_TITLE,	"Rendering",		NULL,		0,					0,					(APTR)Menu_Ignore,
	NM_ITEM,	"Double Strength",	"D",		CHECKIT|MENUTOGGLE,	0,					(APTR)Menu_Double,
	NM_TITLE,	"MIDI",				NULL,		0,					0,					(APTR)Menu_Ignore,
	NM_ITEM,	"Release Notes",	"N",		0,					0,					(APTR)Menu_Release,
	NM_END,		NULL,				NULL,		0,					0,					(APTR)Menu_Ignore,
};


/* Interrupt stuff */

struct Interrupt VertBlank =
{   NULL,NULL,NT_INTERRUPT,-60,"VU VBlank",     /* node, pri = -60 */
    NULL,                                       /* data ptr, same as inputevent */
    (void *)VBlankInterface                     /* code ptr */
};

BOOL Watch = FALSE;
ULONG IdleCount = 0;


/*--------------*/
/* Startup code */
/*--------------*/

LONG __saveds mymain(void)
{
	LONG ReturnCode;
	
	struct	Process *MyProc;
	
	SysBase = *((struct ExecBase**)(0x4));
	
	MyProc = (struct Process*) FindTask(NULL);
	
	if (!MyProc->pr_CLI)
	{
		WBMode = TRUE;
		ReturnCode = WBInterface(MyProc);
	}
	else
	{
		ReturnCode = ShellInterface();
	}
	
	return(ReturnCode);
}



/*---------------------*/
/* Workbench Interface */
/*---------------------*/

LONG WBInterface(struct Process *MyProc)
{
	struct WBStartup *wbmsg;
	
	LONG ReturnCode;
	
	WaitPort(&MyProc->pr_MsgPort);
	
	if (wbmsg = (struct WBStartup*)GetMsg(&MyProc->pr_MsgPort))
	{
		if (OpenLibs())
		{
			ReturnCode = Fireworks("out.0", 100, 200, 256, 256, "PROGDIR:Backdrop.Pic");
			
			CloseLibs();
		}
		Forbid();
		ReplyMsg((struct Message*)wbmsg);
	}
	return(ReturnCode);
}



/*-----------------*/
/* Shell Interface */
/*-----------------*/

LONG  ShellInterface(void)
{
	LONG ReturnCode = RETURN_ERROR;
	
	UBYTE *linkname = "out.0";
	UBYTE *image = "PROGDIR:Backdrop.Pic";
	
	LONG WinX = 100;
	LONG WinY = 200;
	LONG WinW = 256;
	LONG WinH = 256;
	
	if (OpenLibs())
	{
		/* CLI argument parsing */
		
		struct	ArgArray
		{
			UBYTE *aa_Link;
			ULONG *aa_WinX;
			ULONG *aa_WinY;
			ULONG *aa_WinW;
			ULONG *aa_WinH;
			UBYTE *aa_Image;
		} AA = {NULL, NULL, NULL, NULL, NULL, NULL};
		
		static UBYTE	*Template = "LINK/K,WINX/K/N,WINY/K/N,WINW/K/N,WINH/K/N,IMAGE/K";
		struct RDArgs *RDArgs;
		
		ReturnCode = RETURN_FAIL;
		
		if (RDArgs=ReadArgs(Template, (LONG *)&AA, 0))
		{
			if (AA.aa_Link) linkname = AA.aa_Link;
			if (AA.aa_WinX) WinX= *AA.aa_WinX;
			if (AA.aa_WinY) WinY= *AA.aa_WinY;
			if (AA.aa_WinW) WinW= *AA.aa_WinW;
			if (AA.aa_WinH) WinH= *AA.aa_WinH;
			if (AA.aa_Image) image=AA.aa_Image;
			
			ReturnCode = Fireworks(linkname, WinX, WinY, WinW, WinH, image);
			
			FreeArgs(RDArgs);
		}
		else
		{
			PrintFault(IoErr(),"vu");
		}
		CloseLibs();
	}
	return(ReturnCode);
}



/*----------------------------*/
/* MIDI/Window initialisation */
/*----------------------------*/

LONG Fireworks(UBYTE *linkname, LONG WinX, LONG WinY, LONG WinW, LONG WinH, UBYTE *Image)
{
	LONG ReturnCode = RETURN_FAIL;
	
	struct Screen *scr;
	APTR VisualInfo;
	struct Window *Window;
	struct Menu   *Menu;
	
	struct timerequest *treq;
	
	WORD ww, wh;
	WORD minw, minh;
	WORD maxw, maxh;
	
	Object *dto = NULL;
	struct BitMap *BGBitMap = NULL;
	struct BitMap *PaintBitMap = NULL;
	struct RastPort PaintRP;
	struct Layer_Info *LInfo = NULL;
	struct Layer *PaintLayer = NULL;
	UWORD i;
	LONG PenArray[NUMPENS];
	
	APTR NotePool;
	
	struct MidiNode *midi;
	struct MidiLink *link;
	UBYTE buffer[80];
	
	BYTE OldPri;
	
	MyTask = FindTask(NULL);
	
	for (i=0; i<NUMPENS; i++) PenArray[i] = -1L;
	
	if (scr = LockPubScreen(NULL))
	{
		if (!(VisualInfo = GetVisualInfo(scr, TAG_DONE)))
		{
			Message("No visual info!",NULL);
		}
		else
		{
			ww = WinW;
			wh = WinH;
			
			if (!(WindowLayout(scr, &ww, &wh, &minw, &minh, &maxw, &maxh )))
			{
				Message("Couldn't layout window!",NULL);
			}
			else
			{
				if (!(Window = OpenWindowTags( NULL,
						WA_PubScreen, scr,
						WA_Title, "Fireworks",
						WA_Left, WinX,
						WA_Top, WinY,
						WA_InnerWidth, (ULONG)ww,
						WA_InnerHeight, (ULONG)wh,
						WA_GimmeZeroZero, TRUE,
						WA_DepthGadget, TRUE,
						WA_SizeGadget, TRUE,
						WA_SizeBBottom, TRUE,
						WA_CloseGadget, TRUE,
						WA_DragBar, TRUE,
						WA_Activate, TRUE,
						WA_IDCMP, IDCMP_CLOSEWINDOW|IDCMP_NEWSIZE|IDCMP_MENUPICK|IDCMP_REFRESHWINDOW,
						WA_SimpleRefresh, TRUE,
						WA_NewLookMenus, TRUE,
						TAG_DONE ) ))
				{
					Message("Couldn't open window!",NULL);
				}
				else
				{
					ww = Window->Width  - Window->BorderLeft-Window->BorderRight ;
					wh = Window->Height - Window->BorderTop -Window->BorderBottom;
					
					Window->MinWidth  = minw + Window->BorderLeft+Window->BorderRight;
					Window->MinHeight = minh + Window->BorderTop +Window->BorderBottom;
					Window->MaxWidth  = maxw + Window->BorderLeft+Window->BorderRight;
					Window->MaxHeight = maxh + Window->BorderTop +Window->BorderBottom;
					
					if (!(InitOrUpdateGraphics(Window, ww, wh, PenArray, &BGBitMap, &PaintBitMap, &PaintRP, &LInfo, &PaintLayer, Image, &dto)))
					{
						Message("Failed to initialize graphics!", NULL);
					}
					else
					{
						BltBitMapRastPort(PaintBitMap, 0, 0, Window->RPort, 0, 0, ww, wh, 0xc0);
						
						if (Image)
						{
							LoadImage(Window, ww, wh, PenArray, Image, &dto);
							if (dto)
							{
								InitOrUpdateGraphics(Window, ww, wh, PenArray, &BGBitMap, &PaintBitMap, &PaintRP, &LInfo, &PaintLayer, Image, &dto);
								BltBitMapRastPort(PaintBitMap, 0, 0, Window->RPort, 0, 0, ww, wh, 0xc0);
							}
						}
						
						if (!(Menu=(struct Menu *)CreateMenus(VUNewMenu, TAG_DONE)))
						{
							Message("Failed to create intuition menu.",NULL);
						}
						else
						{
							LayoutMenus(Menu, VisualInfo, GTMN_NewLookMenus, TRUE, TAG_DONE);
							SetMenuStrip(Window, Menu);
							
							if (!((NotePool = CreatePool(MEMF_ANY,8192,4096))))
							{
								Message("No memory for NotePool!", NULL);
							}
							else
							{
								if (!(midi = CreateMidi(
									MIDI_Name, "VU Meters",
									MIDI_RecvSignal, SIGBREAKB_CTRL_E,
									MIDI_MsgQueue,   500,
									MIDI_ErrFilter, CMEF_All,
									TAG_DONE)))
								{
									Message("Cannot create MIDI port!",NULL);
								}
								else
								{
									if (!(link = AddMidiLink(midi, MLTYPE_Receiver,
										MLINK_Name, "VU Meter Link",
										MLINK_Location, linkname,
										MLINK_EventMask, CMF_Note|CMF_Mode,
										MLINK_Comment,  "Fireworks [Input]",
										TAG_DONE)))
									{
										Message("Cannot create link to MIDI interface '%s'",NULL,linkname);
									}
									else
									{
										SPrintf(buffer, "Fireworks [%s] [idle]", linkname);
										SetWindowTitles(Window, buffer, (UBYTE*) ~0);
										
										if (!(treq = OpenTimer()))
										{
											
										}
										else
										{
											OldPri = MyTask->tc_Node.ln_Pri;
											
											MainLoop(linkname, Window, &ww, &wh, &midi, &link, NotePool, treq, PenArray, &BGBitMap, &PaintBitMap, &PaintRP, &LInfo, &PaintLayer, Image, &dto);
											ReturnCode = RETURN_OK;
											
											SetTaskPri(MyTask, OldPri);
											
											CloseTimer(treq);
											}
											RemoveMidiLink(link);
										}
										DeleteMidi(midi);
									}
								DeletePool(NotePool);
							}
							ClearMenuStrip(Window);
							
							FreeMenus(Menu);
						}
					}
					FreeGraphics(Window, ww, wh, PenArray, &BGBitMap, &PaintBitMap, &LInfo, &PaintLayer, &dto);
					
					CloseWindow(Window);
				}
			}
			FreeVisualInfo(VisualInfo);
		}
		UnlockPubScreen(NULL, scr);
	}
	return(ReturnCode);
}



BOOL WindowLayout(struct Screen *scr, WORD *ww, WORD *wh, WORD *minw, WORD *minh, WORD *maxw, WORD *maxh)
{
	BOOL Success = FALSE;
	
	*minw=64;
	*minh=64;
	
	*maxw=1024;
	*maxh=1024;
	
	if (*ww < *minw) *ww = *minw;
	if (*wh < *minh) *wh = *minh;
	
	if (*ww > *maxw) *ww = *maxw;
	if (*wh > *maxh) *wh = *maxh;
	
	Success=TRUE;
	
	return(Success);
}



/*-----------------*/
/* Main event loop */
/*-----------------*/

UWORD NoteArray[128];
UBYTE ChanUse[16];
struct MinList CrackerList;

#define SCALE 1024

/* the firecrackers */

struct Cracker
{
	struct Node		cr_node;
	UBYTE			cr_chn;			/* 0 - 15  */
	UBYTE			cr_note;		/* 0 - 127 */
	UBYTE			cr_vel;			/* 0 - 127 */
	
	LONG			cr_starttime;		/* time stamp of note on */
	LONG			cr_stoptime;		/* time stamp of note off */
	UWORD			cr_savedlifetime;	/* lifetime when ray hit border */
	
	WORD			cr_fx;			/* -SCALE to SCALE equals -1 to -1 */
	WORD			cr_fy;			/* -SCALE to SCALE equals -1 to -1 */
};


void MainLoop(UBYTE *linkname, struct Window *Window, WORD *ww, WORD *wh, struct MidiNode **midi, struct MidiLink **link, APTR NotePool, struct timerequest *treq, LONG *PenArray, struct BitMap **BGBitMap, struct BitMap **PaintBitMap, struct RastPort *PaintRP, struct Layer_Info **LInfo, struct Layer **PaintLayer, UBYTE *Image, Object **dto)
{
	BYTE NormPri = -10;
	
	UWORD Mask = 0xffff;
	
	BOOL Active;
	ULONG signals;
	ULONG gotsignals;
	struct IntuiMessage *imsg;
	ULONG Cl;
	UWORD Co;
	APTR IA;
	
	ULONG timersig = (1L << treq->tr_node.io_Message.mn_ReplyPort->mp_SigBit);
	BOOL TimerActive = FALSE;
	ULONG curtime;
	
	UWORD i;
	WORD fxtab[128], fytab[128];
	
	WORD centx, centy;
	WORD divx, divy;
	
	struct Cracker *cr, *ncr;
	UWORD lifetime;
	UWORD deathtime;
	WORD sx1, sy1;
	WORD sx2, sy2;
	
	UBYTE Err;
	
	UBYTE NewImage[240];
	UBYTE NewLink[32];
	UBYTE buffer[80];
	
	BOOL Double=FALSE;
	
	Active=TRUE;
	
	memset(NoteArray,0,sizeof(NoteArray));
	memset(ChanUse  ,0,sizeof(ChanUse  ));
	NewList((struct List*)&CrackerList);
	
	Watch = FALSE;
	AddIntServer(INTB_VERTB, &VertBlank);
	
	/* pre-calculate angles of all possible 128 notes */
	for (i = 0 ; i < 128 ; i++)
	{
		float angle = PI - (float)i/127 * PI;
		
		fxtab[i] = SCALE * cos(angle);
		fytab[i] = SCALE * sin(angle);
	}
	
	*ww = Window->Width  - Window->BorderLeft-Window->BorderRight ;
	*wh = Window->Height - Window->BorderTop -Window->BorderBottom;
	
	/* Calculate ray scaling factors */
	divx  = SCALE * MAXCOORD / (*ww);
	divy  = SCALE * MAXCOORD / (*wh);
	centx = CENTER_X * (*ww) / MAXCOORD;
	centy = CENTER_Y * (*wh) / MAXCOORD;
	
	GetTimeDelta();		/* initialize delta counter */
	
	curtime = 0;		/* start at time counter 0 */
	
	signals = (1L<<Window->UserPort->mp_SigBit) | SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_E | timersig;
	
	TimerActive = FALSE;
	
	while(Active)
	{
		gotsignals = Wait(signals);
		
		if (gotsignals & (1L<<Window->UserPort->mp_SigBit))
		{
			while(Active && (imsg=(struct IntuiMessage*)GetMsg(Window->UserPort)))
			{
				Cl=imsg->Class;
				Co=imsg->Code;
				IA=imsg->IAddress;
				ReplyMsg((struct Message*)imsg);
				
				if (Cl==IDCMP_CLOSEWINDOW)
				{
					Active=FALSE;
					break;
				}
				
				if (Cl==IDCMP_NEWSIZE)
				{
					BOOL SaveWatch = Watch;
					Watch = FALSE;
					
					*ww = Window->Width  - Window->BorderLeft-Window->BorderRight ;
					*wh = Window->Height - Window->BorderTop -Window->BorderBottom;
					
					if (!(InitOrUpdateGraphics(Window, *ww, *wh, PenArray, BGBitMap, PaintBitMap, PaintRP, LInfo, PaintLayer, Image, dto)))
					{
						Message("Failed to update graphics!", NULL);
					}
					
					EraseRect(Window->RPort, 0, 0, (*ww)-1, (*wh)-1);
					
					if (*PaintBitMap)
					{
						BltBitMapRastPort(*PaintBitMap, 0, 0, Window->RPort, 0, 0, *ww, *wh, 0xc0);
					}
					
					/* Calculate ray scaling factors */
					divx  = SCALE * MAXCOORD / (*ww);
					divy  = SCALE * MAXCOORD / (*wh);
					centx = CENTER_X * (*ww) / MAXCOORD;
					centy = CENTER_Y * (*wh) / MAXCOORD;
					
					Watch = SaveWatch;
				}
				
				if (Cl==IDCMP_REFRESHWINDOW)
				{
					BeginRefresh( Window );
					
					if (*PaintBitMap)
					{
						BltBitMapRastPort(*PaintBitMap, 0, 0, Window->RPort, 0, 0, *ww, *wh, 0xc0);
					}
					
				    EndRefresh( Window, TRUE );
				}
				
				if (Cl==IDCMP_MENUPICK)
				{
					BOOL SaveWatch = Watch;
					
					struct MenuItem *n;
					ULONG pick;
					
					Watch = FALSE;
					
					while( (Co != MENUNULL) && Active)
					{
						n = ItemAddress( Window->MenuStrip, (ULONG)Co );
						pick = (ULONG) GTMENUITEM_USERDATA( n );
						
						switch(pick)
						{
							case Menu_Link:
							{
								APTR listreq;
								
								if (!(listreq = AllocListRequest(
									LISTREQ_Window, Window,
									LISTREQ_TitleText, "Select input link",
									TAG_DONE )))
								{
									Message("Couldn't allocate a list requester.",NULL);
								}
								else
								{
									if (linkname != NewLink)
									{
										strncpy(NewLink, (*link)->ml_Location->mcl_Node.ln_Name, sizeof(NewLink));
										NewLink[sizeof(NewLink)-1] = 0;
									}
									
									if (SelectCluster( listreq, NewLink, sizeof(NewLink), TAG_DONE ))
									{
										if (strcmp((*link)->ml_Location->mcl_Node.ln_Name,NewLink))
										{
											struct MidiLink *newlink;
											
											if (!(newlink = AddMidiLink(*midi, MLTYPE_Receiver,
												MLINK_Name, "VU Meter Link",
												MLINK_Location, NewLink,
												MLINK_EventMask, CMF_Note|CMF_Mode,
												MLINK_Comment,  "Fireworks [Input]",
												TAG_DONE)))
											{
												Message("Cannot create link to MIDI interface '%s'",NULL,NewLink);
											}
											else
											{
												RemoveMidiLink( *link );
												FlushMidi ( *midi );
												
												*link = newlink;
												linkname = NewLink;
												
												SPrintf(buffer, "Fireworks [%s]%s", linkname, TimerActive ? "" : " [idle]");
												SetWindowTitles(Window, buffer, (UBYTE*) ~0);
												
												memset(NoteArray,0,sizeof(NoteArray));
												memset(ChanUse  ,0,sizeof(ChanUse  ));
												
												for ( cr = (struct Cracker*) CrackerList.mlh_Head ;
													  ncr = (struct Cracker*) cr->cr_node.ln_Succ ;
													  cr = ncr )
												{
													Remove((struct Node*)cr);
													FreePooled(NotePool,cr,sizeof(struct Cracker));
												}
											}
										}
									}
									FreeListRequest(listreq);
								}
							}
							break;
							
							case Menu_Save:
							{
								Message("Save Config\nIsn't implemented... eeeh!","Damn");
							}
							break;
							
							case Menu_About:
							{
								Message("Fireworks\n 1998 by Christian Buchner\nflowerp@eikon.e-technik.tu-muenchen.de","Cool");
							}
							break;
							
							case Menu_Quit:
							{
								Active = FALSE;
							}
							break;
							
							case Menu_Kill:
							{
								if (*dto)
								{
									SetAPen(Window->RPort, PenArray[Backgroundpen]);
									RectFill(Window->RPort, 0, 0, (*ww)-1, (*wh)-1);
									
									DisposeDTObject(*dto);
									*dto = NULL;
									
									InitOrUpdateGraphics(Window, *ww, *wh, PenArray, BGBitMap, PaintBitMap, PaintRP, LInfo, PaintLayer, Image, dto);
									BltBitMapRastPort(*PaintBitMap, 0, 0, Window->RPort, 0, 0, *ww, *wh, 0xc0);
								}
							}
							break;
							
							case Menu_Load:
							{
								if (Image != NewImage)
								{
									strncpy(NewImage, Image, sizeof(NewImage));
									NewImage[sizeof(NewImage)-1] = 0;
								}
								
								if (SelectImage(Window, NewImage, sizeof(NewImage)))
								{
									Image = NewImage;
									LoadImage(Window, *ww, *wh, PenArray, Image, dto);
									if (*dto)
									{
										InitOrUpdateGraphics(Window, *ww, *wh, PenArray, BGBitMap, PaintBitMap, PaintRP, LInfo, PaintLayer, Image, dto);
										BltBitMapRastPort(*PaintBitMap, 0, 0, Window->RPort, 0, 0, *ww, *wh, 0xc0);
									}
								}
							}
							break;
							
							case Menu_Release:
							{
								memset(NoteArray,0,sizeof(NoteArray));
								memset(ChanUse  ,0,sizeof(ChanUse  ));
								
								curtime += GetTimeDelta();
								
								for ( cr = (struct Cracker*) CrackerList.mlh_Head ;
									  ncr = (struct Cracker*) cr->cr_node.ln_Succ ;
									  cr = ncr )
								{
									if (cr->cr_stoptime == -1) cr->cr_stoptime = curtime;
								}
							}
							break;
							
							case Menu_Double:
							{
								if (n->Flags & CHECKED)
									Double = TRUE;
								else
									Double = FALSE;
							}
							break;
						}
						
						Co = n->NextSelect;
					}
					
					Watch = SaveWatch;
				}
			}
		}
		
		if (Active && (gotsignals & SIGBREAKF_CTRL_E))
		{
			MidiMsg msg;
			
			curtime += GetTimeDelta();
			
			while (GetMidi(*midi,&msg))
			{
				if (noteon(&msg))
				{
					UBYTE chn  = msg.mm_Status & MS_ChanBits;
					UWORD note = msg.mm_Data1;
					UWORD cmsk = 1<<chn;
					
					// if ((Mask & cmsk) && !((NoteArray[note] & Mask)))
					// {
					// }
					
					if (!(NoteArray[note] & cmsk))
					{
						/* Channel wird benutzt */
						ChanUse[chn]++;
						
						NoteArray[note] |= cmsk;
						
						/* Note angeschlagen */
						
						if (cr = AllocPooled(NotePool, sizeof(struct Cracker)))
						{
							cr->cr_node.ln_Pri = chn;
							
							cr->cr_chn  = chn;
							cr->cr_note = note;
							cr->cr_vel  = msg.mm_Data2;
							
							cr->cr_starttime = curtime;
							cr->cr_stoptime  = -1;
							cr->cr_savedlifetime = 0xffff;
							
							cr->cr_fx = fxtab[note] * cr->cr_vel / 127;
							cr->cr_fy = fytab[note] * cr->cr_vel / 127;
							
							/* sorted by channel num to speed up painting */
							Enqueue((struct List*)&CrackerList, (struct Node*)cr);
						}
					}
					
				}
				else
				{
					if (noteoff(&msg))
					{
						UBYTE chn  = msg.mm_Status & MS_ChanBits;
						UWORD note = msg.mm_Data1;
						UWORD cmsk = 1<<chn;
						
						if (NoteArray[note] & cmsk)
						{
							/* Channel wird nicht mehr benutzt */
							ChanUse[chn]--;
							
							NoteArray[note] &= (~cmsk);
						
							/* Note losgelassen */
							
							for ( cr = (struct Cracker*) CrackerList.mlh_Head ;
								  ncr = (struct Cracker*) cr->cr_node.ln_Succ ;
								  cr = ncr )
							{
								if ( (cr->cr_chn  == chn ) &&
									 (cr->cr_note == note) &&
									 (cr->cr_stoptime == -1) )
								{
									cr->cr_stoptime = curtime;
									break;
								}
							}
						}
						
						// if ((Mask & cmsk) && (!(NoteArray[note] & Mask)))
						// {
						// }
					}
					else
					{
						if (modemsg(&msg))
						{
							UBYTE chn  = msg.mm_Status & MS_ChanBits;
							UBYTE mode = msg.mm_Data1;
							UWORD note;
							UWORD cmsk = 1<<chn;
							
							if (mode == MM_AllOff)
							{
								for (note=0;note<128;note++)
								{
									if (NoteArray[note] & cmsk)
									{
										/* Channel wird nicht mehr benuzt */
										ChanUse[chn]--;
										
										/* Note losgelassen */
										
										for ( cr = (struct Cracker*) CrackerList.mlh_Head ;
											  ncr = (struct Cracker*) cr->cr_node.ln_Succ ;
											  cr = ncr )
										{
											if ( (cr->cr_chn  == chn ) &&
												 (cr->cr_note == note) &&
												 (cr->cr_stoptime == -1) )
											{
												cr->cr_stoptime = curtime;
												break;
											}
										}
										
										NoteArray[note] &= (~cmsk);
									}
									
									// if ((Mask & cmsk) && (!(NoteArray[note] & Mask)))
									// {
									// }
								}
							}
						}
					}
				}
			}
			
			if (Err = GetMidiErr(*midi))
			{
				if (Err & CMEF_MsgErr)			Message("MIDI Error: MsgErr!",NULL);
				if (Err & CMEF_BufferFull)		Message("MIDI Error: BufferFull!",NULL);
				if (Err & CMEF_SysExFull)		Message("MIDI Error: SysExFull!",NULL);
				if (Err & CMEF_ParseMem)		Message("MIDI Error: ParseMem!",NULL);
				if (Err & CMEF_RecvErr)			Message("MIDI Error: RecvErr!",NULL);
				if (Err & CMEF_RecvOverflow)	Message("MIDI Error: RecvOverflow!",NULL);
				if (Err & CMEF_SysExTooBig)		Message("MIDI Error: SysExTooBig!",NULL);
			}
			
			/* (re-)start timer if it isn't running */
			if (!TimerActive)
			{
				/* but only if there is something to display */
				if (!IsListEmpty((struct List*)&CrackerList))
				{
					treq->tr_node.io_Command = TR_ADDREQUEST;
					treq->tr_time.tv_secs  = 0;
					treq->tr_time.tv_micro = 1000000 / FPS;
					SendIO((struct IORequest*)treq);
					TimerActive = TRUE;
					Watch = TRUE;
					
					SPrintf(buffer, "Fireworks [%s]", linkname);
					SetWindowTitles(Window, buffer, (UBYTE*) ~0);
				}
			}
		}
		
		if (Active && (gotsignals & timersig))
		{
			IdleCount = 0;
			
			if (!IsListEmpty((struct List*)&CrackerList))
			{
				treq->tr_node.io_Command = TR_ADDREQUEST;
				treq->tr_time.tv_secs  = 0;
				treq->tr_time.tv_micro = 1000000 / FPS;
				SendIO((struct IORequest*)treq);
			}
			else
			{
				UBYTE buffer[80];
				
				TimerActive = FALSE;
				Watch = FALSE;
				
				SPrintf(buffer, "Fireworks [%s] [idle]", linkname);
				SetWindowTitles(Window, buffer, (UBYTE*) ~0);
			}
			
			curtime += GetTimeDelta();
			
			if ((*BGBitMap) && (*PaintBitMap))
			{
				LONG lastpen = -1, newpen;
				
				BltBitMapRastPort(*BGBitMap, 0, 0, PaintRP, 0, 0, *ww, *wh, 0xc0);
				
				for ( cr = (struct Cracker*) CrackerList.mlh_Head ;
					  ncr = (struct Cracker*) cr->cr_node.ln_Succ ;
					  cr = ncr )
				{
					/* get the ray's lifetime information */
					
					if (cr->cr_savedlifetime == 0xffff)
						lifetime = (UWORD)(curtime - cr->cr_starttime);
					else
						lifetime = cr->cr_savedlifetime;
					
					/* calculate ray starting points */
					sx1 = centx + cr->cr_fx *  lifetime / divx;
					sy1 = centy - cr->cr_fy *  lifetime / divy;
					
					if (cr->cr_savedlifetime == 0xffff)
					{
						if ((sx1 > *ww) || (sy1 > *wh) || (sx1 < 0) || (sy1 < 0))
						{
							/* Start of ray hit window borders */
							/* save the time this happened to prevent */
							/* the ray from growing longer than */
							/* graphic's clipping can handle safely */
							cr->cr_savedlifetime = lifetime;
						}
					}
					
					/* has ray already been stopped by a note off? */
					if (cr->cr_stoptime == -1)
					{
						/* no, still emmitting from the centre */
						sx2 = centx;
						sy2 = centy;
					}
					else
					{
						/* get the time when the note off occured */
						deathtime = (UWORD)(curtime - cr->cr_stoptime);
						
						/* is the ray just a dot? */
						if (lifetime == deathtime)
						{
							sx2 = sx1;
							sy2 = sy1;
						}
						else
						{
							/* calculate ray ending points */
							sy2 = centy - cr->cr_fy * deathtime / divy;
							sx2 = centx + cr->cr_fx * deathtime / divx;
						}
					}
					
					/* see if the ray has disappeared */
					/* or exceeded his maximum lifetime of 60 seconds */
					if ((sx2 > *ww) || (sy2 > *wh) || (sx2 < 0) || (sy2 < 0) ||
						lifetime  > 60000 )
					{
						Remove((struct Node*)cr);
						FreePooled(NotePool,cr,sizeof(struct Cracker));
					}
					else
					{
						/* display this channel ? */
						if ((1<<cr->cr_chn) & Mask)
						{
							/* change pen if necessary */
							newpen = PenArray[Channelpens+cr->cr_chn];
							if (lastpen != newpen)
							{
								SetAPen(PaintRP, newpen);
								lastpen = newpen;
							}
							
							/* draw the ray */
							Move(PaintRP, sx1, sy1);
							Draw(PaintRP, sx2, sy2);
							
							if (Double)
							{
								WORD dx, dy;
								dx = sx2 - sx1; if (dx<0) dx=-dx;
								dy = sy2 - sy1; if (dy<0) dy=-dy;
								
								if (dy >= dx)
								{
									Move(PaintRP, sx1+1, sy1);
									Draw(PaintRP, sx2+1, sy2);
								}
								else
								{
									Move(PaintRP, sx1, sy1+1);
									Draw(PaintRP, sx2, sy2+1);
								}
							}
						}
					}
				}
				
				BltBitMapRastPort(*PaintBitMap, 0, 0, Window->RPort, 0, 0, *ww, *wh, 0xc0);
			}
			/* reset task priority if it has
			   been modified by the watchdog */
			if (MyTask->tc_Node.ln_Pri != NormPri)
			{
				SetTaskPri(MyTask, NormPri);
			}
		}
		
		if (gotsignals & SIGBREAKF_CTRL_C)
		{
			Active = FALSE;
			Watch = FALSE;
		}
	}
	
	RemIntServer(INTB_VERTB, &VertBlank);
	
	if (TimerActive)
	{
		AbortIO((struct IORequest*)treq);
		WaitIO((struct IORequest*)treq);
	}
	
	for ( cr = (struct Cracker*) CrackerList.mlh_Head ;
		  ncr = (struct Cracker*) cr->cr_node.ln_Succ ;
		  cr = ncr )
	{
		Remove((struct Node*)cr);
		FreePooled(NotePool,cr,sizeof(struct Cracker));
	}
}



/*-------------------*/
/* Interrupt handler */
/*-------------------*/

int __interrupt __saveds VBlankInterface( void )
{
	if (Watch)
	{
		if (IdleCount++ > 25)
		{
			SetTaskPri(MyTask, 10);
			IdleCount = 0;
		}
	}
	return 0;                      /* server chain continues */
}



/*----------------------------*/
/* Initialize/Update Graphics */
/*----------------------------*/

BOOL InitOrUpdateGraphics(struct Window *Window, WORD ww, WORD wh, LONG *PenArray, struct BitMap **BGBitMap, struct BitMap **PaintBitMap, struct RastPort *PaintRP, struct Layer_Info **LInfo, struct Layer **PaintLayer, UBYTE *Image, Object **dto)
{
	BOOL Success = FALSE;
	
	if (*PaintLayer)
	{
		DeleteLayer(0, *PaintLayer);
		*PaintLayer = NULL;
	}
	
	if (*BGBitMap)
	{
		FreeBitMap(*BGBitMap);
		*BGBitMap = NULL;
	}
	
	if (*PaintBitMap)
	{
		FreeBitMap(*PaintBitMap);
		*PaintBitMap = NULL;
	}
	
	if (Window)
	{
		UWORD i;
		struct ColorMap *cmap = Window->WScreen->ViewPort.ColorMap;
		
		struct BitMap *friend = Window->RPort->BitMap;
		ULONG depth = GetBitMapAttr(friend, BMA_DEPTH);
		ULONG flags = GetBitMapAttr(friend, BMA_FLAGS);
		
		struct RastPort BGRP;
		
		SetPointer(Window, WaitPointer, 16, 16, -6, 0);
		
		for (i = 0 ; i < NUMPENS ; i++)
		{
			if (PenArray[i] == -1L)
			{
				PenArray[i] = ObtainBestPen(
					cmap,
					(ULONG)PenColors[i][0]<<24, 
					(ULONG)PenColors[i][1]<<24, 
					(ULONG)PenColors[i][2]<<24,
					OBP_Precision, PRECISION_IMAGE,
					TAG_DONE );
			}
		}
		
		/* Cybergraphics kludge */
		if ( !(flags & BMF_STANDARD) ) flags |= BMF_MINPLANES;
		
		if (*BGBitMap = AllocBitMap(ww, wh, depth, flags | BMF_CLEAR, friend))
		{
			struct BitMap *ImageBM   = NULL;
			struct BitMapHeader *bmh = NULL;
			
			InitRastPort(&BGRP);
			BGRP.BitMap = *BGBitMap;
			
			SetAPen(&BGRP, PenArray[Backgroundpen]);
			
			if (*dto)
			{
				GetDTAttrs(*dto, PDTA_DestBitMap, &ImageBM,
								 PDTA_BitMapHeader, &bmh,
							TAG_DONE);
			}
			
			if ((!ImageBM) || (!bmh))
			{
				RectFill(&BGRP, 0, 0, ww-1, wh-1);
			}
			else
			{
				struct BitScaleArgs bsa;
				
				UWORD width = bmh->bmh_Width;
				UWORD height= bmh->bmh_Height;
				
				bsa.bsa_SrcX		= 0;
				bsa.bsa_SrcY		= 0;
				bsa.bsa_SrcWidth	= width;
				bsa.bsa_SrcHeight	= height;
				bsa.bsa_XSrcFactor	= width;
				bsa.bsa_YSrcFactor	= height;
				bsa.bsa_DestX		= 0;
				bsa.bsa_DestY		= 0;
				bsa.bsa_DestWidth	= ww;
				bsa.bsa_DestHeight	= wh;
				bsa.bsa_XDestFactor	= ww;
				bsa.bsa_YDestFactor	= wh;
				bsa.bsa_SrcBitMap	= ImageBM;
				bsa.bsa_DestBitMap	= *BGBitMap;
				bsa.bsa_Flags		= 0;
				bsa.bsa_XDDA		= 0;
				bsa.bsa_YDDA		= 0;
				bsa.bsa_Reserved1	= 0;
				bsa.bsa_Reserved2	= 0;
				
				BitMapScale(&bsa);
			}
			
			if ( *PaintBitMap = AllocBitMap(ww, wh, depth, flags | BMF_CLEAR, friend))
			{
				InitRastPort(PaintRP);
				PaintRP->BitMap = *PaintBitMap;
				
				if (!(*LInfo)) *LInfo = NewLayerInfo();
				
				if (*LInfo)
				{
					if (*PaintLayer = CreateUpfrontLayer(*LInfo, *PaintBitMap, 0, 0, ww-1, wh-1, LAYERSMART, NULL))
					{
						PaintRP->Layer = *PaintLayer;
						
						BltBitMapRastPort(*BGBitMap, 0, 0, PaintRP, 0, 0, ww, wh, 0xc0);
						
						Success = TRUE;
					}
				}
			}
		}
		ClearPointer(Window);
	}
	
	if (!Success)
	{
		if (*PaintLayer)
		{
			DeleteLayer(0, *PaintLayer);
			*PaintLayer = NULL;
		}
		
		if (*BGBitMap)
		{
			FreeBitMap(*BGBitMap);
			*BGBitMap = NULL;
		}
		
		if (*PaintBitMap)
		{
			FreeBitMap(*PaintBitMap);
			*PaintBitMap = NULL;
		}
	}
	
	return(Success);
}



/*-------------------------*/
/* Free allocated Graphics */
/*-------------------------*/

void FreeGraphics(struct Window *Window, UWORD ww, UWORD wh, LONG *PenArray, struct BitMap **BGBitMap, struct BitMap **PaintBitMap, struct Layer_Info **LInfo, struct Layer **PaintLayer, Object **dto)
{
	UWORD i;
	
	if (*dto)
	{
		SetAPen(Window->RPort, PenArray[Backgroundpen]);
		RectFill(Window->RPort, 0, 0, ww-1, wh-1);
		
		DisposeDTObject(*dto);
		*dto = NULL;
	}
	
	if (*PaintLayer)
	{
		DeleteLayer(0, *PaintLayer);
		*PaintLayer = NULL;
	}
	
	if (*LInfo)
	{
		DisposeLayerInfo(*LInfo);
		*LInfo = NULL;
	}
	
	if (*BGBitMap)
	{
		FreeBitMap(*BGBitMap);
		BGBitMap = NULL;
	}
	
	if (*PaintBitMap)
	{
		FreeBitMap(*PaintBitMap);
		*PaintBitMap = NULL;
	}
	
	if (Window)
	{
		struct ColorMap *cmap = Window->WScreen->ViewPort.ColorMap;
		
		for (i = 0 ; i < NUMPENS ; i++)
		{
			if (PenArray[i] != -1L)
			{
				ReleasePen( cmap, PenArray[i] );
			}
		}
	}
}



/*----------------------------------*/
/* Load image using datatypes (V43) */
/*----------------------------------*/

void LoadImage(struct Window *Window, UWORD ww, UWORD wh, LONG *PenArray, UBYTE *Image, Object **dto)
{
	if (*dto)
	{
		SetAPen(Window->RPort, PenArray[Backgroundpen]);
		RectFill(Window->RPort, 0, 0, ww-1, wh-1);
		
		DisposeDTObject(*dto);
		*dto = NULL;
	}
	
	SetPointer(Window, WaitPointer, 16, 16, -6, 0);
	
	if (*dto = (Object *) NewDTObject(Image,
		DTA_GroupID, GID_PICTURE, 
		PDTA_DestMode, MODE_V43,
		PDTA_Remap, TRUE,
		PDTA_Screen, Window->WScreen,
		PDTA_FreeSourceBitMap, TRUE,
		TAG_DONE) )
	{
		SetDTAttrs(*dto, NULL, NULL,
					PDTA_Remap, TRUE,
					PDTA_Screen, Window->WScreen,
					PDTA_FreeSourceBitMap, TRUE,
					PDTA_UseFriendBitMap, TRUE,
					TAG_DONE );
		
		DoMethod( *dto, DTM_PROCLAYOUT, NULL, TRUE );
	}
	else
	{
		LONG errnum=IoErr();
		UBYTE errstring[80];
		
		if (errnum>=DTERROR_UNKNOWN_DATATYPE)
		{
			SPrintf(errstring,GetDTString(errnum),Image);
		}
		else
		{
			Fault(errnum,NULL,errstring,sizeof(errstring));
		}
		Message("%s: %s",NULL,Image,errstring);
	}
	
	ClearPointer(Window);
}



/*-------------------------------------------*/
/* Datatypes hook function for ASL requester */
/*-------------------------------------------*/

ULONG __asm __saveds Filter (register __a0 struct Hook *h, register __a2 struct FileRequester *fr, register __a1 struct AnchorPath *ap)
{
	struct DataType *dtn;
	ULONG use = FALSE;
	UBYTE buffer[300];
	BPTR lock;
	
	if (ap->ap_Info.fib_DirEntryType > 0)
	{
		use = TRUE;
	}
	else
	{
		strncpy (buffer, fr->fr_Drawer, sizeof (buffer));
		AddPart (buffer, ap->ap_Info.fib_FileName, sizeof (buffer));
		if (lock = Lock (buffer, ACCESS_READ))
		{
			if (dtn = ObtainDataTypeA (DTST_FILE, (APTR) lock, NULL))
			{
				if (dtn->dtn_Header->dth_GroupID == GID_PICTURE)
					use = TRUE;
				
				ReleaseDataType (dtn);
			}
			UnLock (lock);
		}
	}
	return (use);
}


/*--------------------------------------*/
/* Select image to load (ASL requester) */
/*--------------------------------------*/

BOOL SelectImage(struct Window *Window, UBYTE *filebuffer, ULONG MaxSize)
{
	BOOL Selected=FALSE;
	
	UBYTE initialdrawer[200];
	UBYTE initialfile  [40];
	
	struct FileRequester *fr;
	struct Hook filter = {NULL, NULL, NULL, NULL, NULL};
	UBYTE save, *ptr;
	
	filter.h_Entry = (HOOKFUNC)Filter;
	
	ptr = PathPart(filebuffer);
	save = *ptr; *ptr = 0;
	strncpy (initialdrawer,filebuffer,sizeof(initialdrawer));
	initialdrawer[sizeof(initialdrawer)-1]=0;
	*ptr = save;
	strncpy (initialfile, FilePart(filebuffer), sizeof(initialfile));
	initialfile[sizeof(initialfile)-1]=0;
	
	if (!(fr = AllocAslRequestTags (ASL_FileRequest,
					ASLFR_Window, Window,
					ASLFR_TitleText,	"Select Picture to Open",
					ASLFR_PositiveText,	"Open",
					ASLFR_RejectIcons,	TRUE,
					ASLFR_FilterFunc,	&filter,
					ASLFR_InitialDrawer, initialdrawer,
					ASLFR_InitialFile, initialfile,
					TAG_DONE)))
	{
		Message("Couldn't allocate asl requester", NULL);
	}
	else
	{
		Selected = AslRequestTags(fr, TAG_DONE);
		
		if (Selected)
		{
			strncpy (filebuffer, fr->fr_Drawer, MaxSize);
			AddPart (filebuffer, fr->fr_File,   MaxSize);
		}
		
		FreeAslRequest(fr);
	}
	
	return(Selected);
}



/*----------------*/
/* Open Libraries */
/*----------------*/

BOOL OpenLibs(void)
{
	BOOL Success=FALSE;
	
	SysBase = *((struct ExecBase**)(0x4));
	
	if (DOSBase=(struct DosLibrary*)OpenLibrary("dos.library",39L))
	{
		if (GfxBase=(struct GfxBase*)OpenLibrary("graphics.library",39L))
		{
			if (LayersBase=OpenLibrary("layers.library",39L))
			{
				if (IntuitionBase=(struct IntuitionBase*)OpenLibrary("intuition.library",39L))
				{
					if (GadToolsBase=OpenLibrary("gadtools.library", 39L))
					{
						if (UtilityBase=(struct Library*)OpenLibrary("utility.library",39L))
						{
							if (!(DatatypesBase=(struct Library*)OpenLibrary("datatypes.library",39L)))
							{
								Message("This program requires datatypes.library!",NULL);
							}
							else
							{
								if (!(AslBase=(struct Library*)OpenLibrary("asl.library",39L)))
								{
									Message("This program requires asl.library!",NULL);
								}
								else
								{
									if (!(CamdBase=OpenLibrary("camd.library",0L)))
									{
										Message("This program requires camd.library!",NULL);
									}
									else
									{
										if ( __fpinit() )
										{
											Message("Unable to initialize floating point!",NULL);
										}
										else
										{
											FPBase=TRUE;
											
											Success=TRUE;
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}
	if (!Success) CloseLibs();
	
	return(Success);
}


/*-----------------*/
/* Close Libraries */
/*-----------------*/

void CloseLibs(void)
{
	if (FPBase)
	{
		__fpterm();
		FPBase=FALSE;
	}
	
	if (CamdBase)
	{
		CloseLibrary(CamdBase);
		CamdBase=NULL;
	}
	
	if (AslBase)
	{
		CloseLibrary(AslBase);
		AslBase=NULL;
	}
	
	if (DatatypesBase)
	{
		CloseLibrary(DatatypesBase);
		DatatypesBase=NULL;
	}
	
	if (UtilityBase)
	{
		CloseLibrary(UtilityBase);
		UtilityBase=NULL;
	}
	
	if (GadToolsBase)
	{
		CloseLibrary(GadToolsBase);
		GadToolsBase=NULL;
	}
	
	if (IntuitionBase)
	{
		CloseLibrary((struct Library*)IntuitionBase);
		IntuitionBase=NULL;
	}
	
	if (LayersBase)
	{
		CloseLibrary((struct Library*)LayersBase);
		LayersBase=NULL;
	}
	
	if (GfxBase)
	{
		CloseLibrary((struct Library*)GfxBase);
		GfxBase=NULL;
	}
	
	if (DOSBase)
	{
		CloseLibrary((struct Library*)DOSBase);
		DOSBase=NULL;
	}
}


/*-------------------*/
/* Open timer device */
/*-------------------*/

struct timerequest *OpenTimer(void)
{
	BOOL Success = FALSE;
	
	struct MsgPort *reply = NULL;
	struct timerequest *treq = NULL;
	
	if (!(reply = CreateMsgPort()))
	{
		Message("Unable to create timer port!", NULL);
	}
	else
	{
		if (!(treq = CreateIORequest(reply, sizeof(struct timerequest))))
		{
			Message("Unable to create timer request!", NULL);
		}
		else
		{
			if (OpenDevice(TIMERNAME,UNIT_VBLANK,(struct IORequest*)treq,0))
			{
				Message("Timer: Cannot open timer.device (UNIT_VBLANK).", NULL);
			}
			else
			{
				TimerBase = (struct Library*)treq->tr_node.io_Device;
				Success = TRUE;
			}
		}
	}
	
	if (!Success)
	{
		if (TimerBase)
		{
			CloseDevice((struct IORequest*)treq);
			TimerBase = NULL;
		}
		
		if (treq)
		{
			DeleteIORequest(treq);
			treq = NULL;
		}
		
		if (reply)
		{
			DeleteMsgPort(reply);
			reply = NULL;
		}
	}
	
	return(treq);
}



/*--------------------*/
/* Close timer device */
/*--------------------*/

void CloseTimer(struct timerequest *treq)
{
	struct MsgPort *reply = NULL;
	
	if (treq)
	{
		if (treq->tr_node.io_Device)
		{
			CloseDevice((struct IORequest*)treq);
			TimerBase = NULL;
		}
		
		reply = treq->tr_node.io_Message.mn_ReplyPort;
		
		DeleteIORequest(treq);
		treq = NULL;
	}
	
	if (reply)
	{
		DeleteMsgPort(reply);
		reply = NULL;
	}
}


/*----------------*/
/* Get Time Delta */
/*----------------*/

/* This function returns the time difference in milliseconds */
/* to the time it was called last                            */

struct EClockVal eclock;

ULONG GetTimeDelta(void)
{
	// ULONG old_hi = eclock.ev_hi;
	ULONG old_lo = eclock.ev_lo;
	ULONG E_Freq = ReadEClock(&eclock);
	
	ULONG delta = 1000 * (eclock.ev_lo - old_lo) / E_Freq;
	
	return(delta);
}



/*****************************************************************************/


/*----------------------------*/
/* Show a message to the user */
/*----------------------------*/

LONG __stdargs Message(UBYTE *Msg,UBYTE *Options,...)
{
	LONG retval;
	
	BOOL req = FALSE;
	
	va_list Arg;
	va_start(Arg,Options);
	
	// if (Options) if (strchr(Options,'|')) req = TRUE;
	if (Options) req = TRUE;
	
	if (IntuitionBase && (WBMode || req))
	{
		struct EasyStruct Req={sizeof(struct EasyStruct),0,"Fireworks",0, NULL};
		
		if (!Options) Options = "I see";
		
		Req.es_TextFormat=Msg;
		Req.es_GadgetFormat=Options;
		
		retval=EasyRequestArgs(NULL,&Req,0,Arg);
	}
	else
	{
		if (DOSBase)
		{
			VPrintf(Msg,Arg);
			Printf("\n");
			
			retval=0;
		}
	}
	va_end(Arg);
	
	return(retval);
}


/*-----------------------*/
/* Standard Exit routine */
/*-----------------------*/

void __stdargs _XCEXIT(LONG lcode)
{
	Message("Task wants to exit, return code %ld\nHolding task!", NULL, lcode);
	Wait(0);
}


/*-------------------*/
/* CAM Library stubs */
/*-------------------*/

struct MidiNode *CreateMidi(Tag tag, ...)
{
	return CreateMidiA((struct TagItem *)&tag );
}

BOOL SetMidiAttrs(struct MidiNode *mi, Tag tag, ...)
{
	return SetMidiAttrsA(mi, (struct TagItem *)&tag );
}

struct MidiLink *AddMidiLink(struct MidiNode *mi, LONG type, Tag tag, ...)
{
	return AddMidiLinkA(mi, type, (struct TagItem *)&tag );
}

BOOL SetMidiLinkAttrs(struct MidiLink *mi, Tag tag, ...)
{
	return SetMidiLinkAttrsA(mi, (struct TagItem *)&tag );
}
